%% 
%
% Leverage optimization example code (ICRA 2015)
%
% Sungjoon Choi (sungjoon.s.choi@gmail.com)
%
ccc;

%% 1. Generate Data (2-dimensional field)
ccc;
stage_struct = init_stage([0 40 0 30], [40 30]);
hyp_se_gen = struct('g2', 1, 'l2', 30, 'w2', 1E-4);
kfun_se_gen = @(x, y)(kernel_se(x, y, hyp_se_gen));
field_struct = gen_gp_field(stage_struct, kfun_se_gen);
meas_struct = get_meas_field(field_struct, stage_struct, 100);
% plot
plot_field_meas(field_struct, meas_struct);
% save('mats/field_meas_struct.mat', 'field_struct', 'meas_struct');

%% 2. Reconstruct the field using GPR 
ccc; 
load('mats/field_meas_struct.mat');
hyp = struct('g2', 1E0, 'l2', 2E1, 'w2', 1E-6);
sgpr_struct = init_lgpr(meas_struct.input, meas_struct.output, @kernel_lev_cos, hyp);
sgpr_result = lgpr(sgpr_struct, field_struct.points);
recon_field_ref = reshape(sgpr_result.mean, field_struct.yres, field_struct.xres);
% plot
str1 = 'Reference Field'; str2 = 'Reconstructed Field';
plot_two_fields(field_struct, field_struct.zdata, recon_field_ref, meas_struct, str1, str2);

%% 3. Change leveage parameter (outlier) and reconstruct field
ccc; rng('shuffle');
load('mats/field_meas_struct.mat');
hyp = struct('g2', 1E0, 'l2', 2E1, 'w2', 1E-6);
sgpr_struct = init_lgpr(meas_struct.input, meas_struct.output, @kernel_lev_cos, hyp);
sgpr_result = lgpr(sgpr_struct, field_struct.points);
recon_field_ref = reshape(sgpr_result.mean, field_struct.yres, field_struct.xres);
% Add outlier by changing lev from +1 to -1
rand_idx = randi([1 meas_struct.nr_data]);
lev_noisy = ones(meas_struct.nr_data, 1);
lev_noisy(rand_idx) = -1;
lgpr_noisy_struct = init_lgpr(meas_struct.input, meas_struct.output, @kernel_lev_cos, hyp, lev_noisy);
lgpr_noisy_result = lgpr(lgpr_noisy_struct, field_struct.points);
recon_field_noisy = reshape(lgpr_noisy_result.mean, field_struct.yres, field_struct.xres);
% plot
plot_ref_recon_fields(field_struct, recon_field_ref, recon_field_noisy, meas_struct, rand_idx);
% save
save_idx = 3;
save_str = sprintf('mats/field_meas_lev_noisy_struct_%02d.mat', save_idx);
% save(save_str, 'field_struct', 'meas_struct', 'lev_noisy', 'rand_idx');

%% 4. Check whether we can find the outlier using Exhaustive search of maximizing the "Marginal Likelihood"
ccc;
% load
load_idx = 3;
load_str = sprintf('mats/field_meas_lev_noisy_struct_%02d.mat', load_idx);
load(load_str);
% Original field (using Reference measurements)
hyp = struct('g2', 1E0, 'l2', 2E1, 'w2', 1E-6);
sgpr_struct = init_lgpr(meas_struct.input, meas_struct.output, @kernel_lev_cos, hyp);
sgpr_result = lgpr(sgpr_struct, field_struct.points);
recon_field_ref = reshape(sgpr_result.mean, field_struct.yres, field_struct.xres);
% Noisy field (with one outlier)
lev_noisy = ones(meas_struct.nr_data, 1);
lev_noisy(rand_idx) = -1;
temp_struct = init_lgpr(meas_struct.input, meas_struct.output, @kernel_lev_cos, hyp, lev_noisy);
temp_result = lgpr(temp_struct, meas_struct.input(rand_idx, 1));
noisy_meas_struct = meas_struct;
noisy_meas_struct.output(rand_idx) = temp_result.mean;
lgpr_noisy_struct = init_lgpr(noisy_meas_struct.input, noisy_meas_struct.output, @kernel_lev_cos, hyp);
lgpr_noisy_result = lgpr(lgpr_noisy_struct, field_struct.points);
recon_field_noisy = reshape(lgpr_noisy_result.mean, field_struct.yres, field_struct.xres);
% Exhaustive search !! 
Lloo_list = zeros(noisy_meas_struct.nr_data, 1);
for i = 1:noisy_meas_struct.nr_data
    lev_mod = ones(noisy_meas_struct.nr_data, 1);
    lev_mod(i) = -1;
    kfun_mod = @(x, y, hyp)(kernel_lev_cos(x, y, lev_mod, lev_mod, hyp));
    % Lloo_noisy_mod = get_Lloo(noisy_meas_struct, kfun_mod, hyp);
    Lloo_noisy_mod = get_Lmarginal(noisy_meas_struct, kfun_mod, hyp);
    Lloo_list(i) = Lloo_noisy_mod;
    % fprintf('%d: Lloo: %.2e \n', i, Lloo_noisy_mod);
end
% We should find 'rand_idx'
[Lloo_opt, idx_opt] = max(Lloo_list);
fprintf('Answer: %d / Solution: %d (It should be the same) \n', idx_opt, rand_idx);
% Reconstruct the field using optimized leverage parameter 
lev_opt = ones(noisy_meas_struct.nr_data, 1);
lev_opt(idx_opt) = -1;
lgpr_struct_opt = init_lgpr(noisy_meas_struct.input, noisy_meas_struct.output, @kernel_lev_cos, hyp, lev_opt);
lgpr_result_opt = lgpr(lgpr_struct_opt, field_struct.points);
recon_field_opt = reshape(lgpr_result_opt.mean, field_struct.yres, field_struct.xres);
% Reconstruction error
recon_ref_error = norm(field_struct.zdata - recon_field_ref);
recon_noisy_error = norm(field_struct.zdata - recon_field_noisy);
recon_opt_error = norm(field_struct.zdata - recon_field_opt);
fprintf('[%d] Ref error: %.2e / Noisy error: %.2e / Optimized error: %.2e (%.1f%% better) \n' ...
    , load_idx, recon_ref_error, recon_noisy_error, recon_opt_error, 100*(recon_noisy_error-recon_opt_error)/recon_noisy_error);
% Plot
plot_ref_recon_opt_fields(field_struct, recon_field_ref, recon_field_noisy, recon_field_opt, noisy_meas_struct, rand_idx, idx_opt);



%%
